# coding: utf-8
"""get list of :class:`objects<supervisely.annotation.obj_class.ObjClass>` from supervisely project"""

from __future__ import annotations
from typing import NamedTuple, List, Dict, Optional

from supervisely.api.module_api import ModuleApi
from supervisely.api.module_api import ApiField
from supervisely.video_annotation.key_id_map import KeyIdMap


class ObjectClassApi(ModuleApi):
    """
    API for working with :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>`. :class:`ObjectClassApi<ObjectClassApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # You can connect to API directly
        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Or you can use API from environment
        os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        project_id = 1951
        obj_class_infos = api.object_class.get_list(project_id)
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple ObjectClassInfo information about ObjectClass.

        :Example:

         .. code-block:: python

            ObjectClassInfo(id=22309,
                            name='lemon',
                            description='',
                            shape='bitmap',
                            color='#51C6AA',
                            settings={},
                            created_at='2021-03-02T10:04:33.973Z',
                            updated_at='2021-03-11T09:37:07.111Z')
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.DESCRIPTION,
            ApiField.SHAPE,
            ApiField.COLOR,
            ApiField.SETTINGS,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **ObjectClassInfo**.
        """
        return "ObjectClassInfo"

    def get_list(
        self, project_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[NamedTuple]:
        """
        List of ObjClasses in the given Project.

        :param project_id: Project ID in which the ObjClasses are located.
        :type project_id: int
        :param filters: List of params to sort output ObjClasses.
        :type filters: List[dict], optional
        :return: List of ObjClasses with information from the given Project. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[NamedTuple]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_id = 1951
            obj_class_infos = api.object_class.get_list(project_id)
            print(obj_class_infos)
            # Output: [ObjectClassInfo(id=22309,
            #                          name='lemon',
            #                          description='',
            #                          shape='bitmap',
            #                          color='#51C6AA',
            #                          settings={},
            #                          created_at='2021-03-02T10:04:33.973Z',
            #                          updated_at='2021-03-11T09:37:07.111Z'),
            #  ObjectClassInfo(id=22310,
            #                  name='kiwi',
            #                  description='',
            #                  shape='bitmap',
            #                  color='#FF0000',
            #                  settings={},
            #                  created_at='2021-03-02T10:04:33.973Z',
            #                  updated_at='2021-03-11T09:37:07.111Z')
            # ]

            obj_class_list = api.object_class.get_list(1951, filters=[{'field': 'name', 'operator': '=', 'value': 'lemon' }])
            print(obj_class_list)
            # Output: [
            #     [
            #         22309,
            #         "lemon",
            #         "",
            #         "bitmap",
            #         "#51C6AA",
            #         {},
            #         "2021-03-02T10:04:33.973Z",
            #         "2021-03-11T09:37:07.111Z"
            #     ]
            # ]
        """
        return self.get_list_all_pages(
            "advanced.object_classes.list",
            {ApiField.PROJECT_ID: project_id, "filter": filters or []},
        )

    def get_name_to_id_map(self, project_id: int) -> Dict[str, int]:
        """
        :param project_id: Project ID in which the ObjClasses are located.
        :type project_id: int
        :return: Dictionary Key ID Map {'key': id}
        :rtype: :class:`KeyIdMap<supervisely.video_annotation.key_id_map.KeyIdMap>`
        :Usage example:

        .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            obj_class_map = api.object_class.get_name_to_id_map(1951)
            print(obj_class_map)
            # Output: {'lemon': 22309, 'kiwi': 22310, 'cucumber': 22379}
        """
        objects_infos = self.get_list(project_id)
        return {object_info.name: object_info.id for object_info in objects_infos}

    def _get_info_by_id(self, id, method, fields=None):
        response = self._get_response_by_id(id, method, id_field=ApiField.ID, fields=fields)
        return (
            self._convert_json_info(response.json(), skip_missing=True)
            if (response is not None)
            else None
        )

    def get_info_by_id(self, id):
        return self._get_info_by_id(
            id,
            "advanced.object_classes.info",
        )

    # def _object_classes_to_json(self, object_classes: KeyIndexedCollection, objclasses_name_id_map=None, project_id=None):
    #     pass #@TODO: implement
    #     # if objclasses_name_id_map is None and project_id is None:
    #     #     raise RuntimeError("Impossible to get ids for projectTags")
    #     # if objclasses_name_id_map is None:
    #     #     objclasses_name_id_map = self.get_name_to_id_map(project_id)
    #     # tags_json = []
    #     # for tag in tags:
    #     #     tag_json = tag.to_json()
    #     #     tag_json[ApiField.TAG_ID] = tag_name_id_map[tag.name]
    #     #     tags_json.append(tag_json)
    #     # return tags_json
    #
    # def append_to_video(self, video_id, tags: KeyIndexedCollection, key_id_map: KeyIdMap = None):
    #     if len(tags) == 0:
    #         return []
    #     video_info = self._api.video.get_info_by_id(video_id)
    #     tags_json = self._tags_to_json(tags, project_id=video_info.project_id)
    #     ids = self.append_to_video_json(video_id, tags_json)
    #     KeyIdMap.add_tags_to(key_id_map, [tag.key() for tag in tags], ids)
    #     return ids
    #
    # def append_to_video_json(self, video_id, tags_json):
    #     if len(tags_json) == 0:
    #         return
    #     response = self._api.post('videos.tags.bulk.add', {ApiField.VIDEO_ID: video_id, ApiField.TAGS: tags_json})
    #     ids = [obj[ApiField.ID] for obj in response.json()]
    #     return ids



# coding: utf-8
from supervisely.video_annotation.video_object_collection import VideoObjectCollection
from supervisely.volume_annotation.volume_object import VolumeObject


class VolumeObjectCollection(VideoObjectCollection):
    item_type = VolumeObject



from typing import List
from supervisely.app.widgets import Widget, Container, generate_id
from supervisely.sly_logger import logger
from supervisely._utils import batched, rand_str
from supervisely.app.widgets.empty.empty import Empty


class Grid(Widget):
    def __init__(
        self,
        widgets: List[Widget],
        columns: int = 1,
        gap: int = 10,
        widget_id: str = None,
    ):
        self._widgets = widgets
        self._columns = columns
        self._gap = gap

        if self._columns < 1:
            raise ValueError(f"columns ({self._columns}) < 1")
        if self._columns > len(self._widgets):
            logger.warn(
                f"Number of columns ({self._columns}) > number of widgets ({len(self._widgets)}). Columns are set to {len(self._widgets)}"
            )
            self._columns = len(self._widgets)

        self._content = None
        if self._columns == 1:
            self._content = Container(
                direction="vertical",
                widgets=self._widgets,
                gap=self._gap,
                widget_id=generate_id(),
            )
        else:
            rows = []
            num_empty = len(self._widgets) % self._columns
            self._widgets.extend([Empty()] * num_empty)
            for batch in batched(self._widgets, batch_size=self._columns):
                rows.append(
                    Container(
                        direction="horizontal",
                        widgets=batch,
                        gap=self._gap,
                        fractions=[1] * len(batch),
                        widget_id=generate_id(),
                    )
                )
            self._content = Container(
                direction="vertical",
                widgets=rows,
                gap=self._gap,
                widget_id=generate_id(),
            )

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return None

    def get_json_state(self):
        return None



import os
import signal
import psutil
import sys
from pathlib import Path

from fastapi import (
    FastAPI,
    Request,
    Response,
    WebSocket,
    WebSocketDisconnect,
    Depends,
    HTTPException,
)

# from supervisely.app.fastapi.request import Request

import jinja2
from fastapi.exception_handlers import http_exception_handler
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from supervisely.app.fastapi.utils import run_sync
from supervisely.app.singleton import Singleton
from supervisely.app.fastapi.templating import Jinja2Templates
from supervisely.app.fastapi.websocket import WebsocketManager
from supervisely.io.fs import mkdir, dir_exists
from supervisely.sly_logger import logger
from supervisely.api.api import SERVER_ADDRESS, API_TOKEN, TASK_ID, Api
from supervisely._utils import is_production, is_development
from async_asgi_testclient import TestClient
from supervisely.app.widgets_context import JinjaWidgets
from supervisely.app.exceptions import DialogWindowBase
import supervisely.io.env as sly_env

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from supervisely.app.widgets import Widget


def create(process_id=None, headless=False) -> FastAPI:
    from supervisely.app import DataJson, StateJson

    app = FastAPI()
    WebsocketManager().set_app(app)

    @app.post("/shutdown")
    async def shutdown_endpoint(request: Request):
        shutdown(process_id)

    if headless is False:

        @app.post("/data")
        async def send_data(request: Request):
            data = DataJson()
            response = JSONResponse(content=dict(data))
            return response

        @app.post("/state")
        async def send_state(request: Request):
            state = StateJson()
            response = JSONResponse(content=dict(state))

            gettrace = getattr(sys, "gettrace", None)
            if (gettrace is not None and gettrace()) or is_development():
                response.headers["x-debug-mode"] = "1"
            return response

        @app.post("/session-info")
        async def send_session_info(request: Request):
            server_address = "/"
            if is_development():
                server_address = sly_env.server_address()
                if server_address is not None:
                    server_address = Api.normalize_server_address(server_address)

            response = JSONResponse(
                content={
                    TASK_ID: os.environ.get(TASK_ID),
                    SERVER_ADDRESS: server_address,
                    API_TOKEN: os.environ.get(API_TOKEN),
                }
            )
            return response

        @app.websocket("/ws")
        async def websocket_endpoint(websocket: WebSocket):
            await WebsocketManager().connect(websocket)
            try:
                while True:
                    data = await websocket.receive_text()
            except WebSocketDisconnect:
                WebsocketManager().disconnect(websocket)

        import supervisely

        app.mount("/css", StaticFiles(directory=supervisely.__path__[0]), name="sly_static")

    return app


def shutdown(process_id=None):
    try:
        logger.info(f"Shutting down [pid argument = {process_id}]...")
        if process_id is None:
            # process_id = psutil.Process(os.getpid()).ppid()
            process_id = os.getpid()
        current_process = psutil.Process(process_id)
        current_process.send_signal(signal.SIGINT)  # emit ctrl + c
    except KeyboardInterrupt:
        logger.info("Application has been shut down successfully")


def enable_hot_reload_on_debug(app: FastAPI):
    templates = Jinja2Templates()
    gettrace = getattr(sys, "gettrace", None)
    if gettrace is None:
        print("Can not detect debug mode, no sys.gettrace")
    elif gettrace():
        import arel

        hot_reload = arel.HotReload(paths=[arel.Path(".")])
        app.add_websocket_route("/hot-reload", route=hot_reload, name="hot-reload")
        app.add_event_handler("startup", hot_reload.startup)
        app.add_event_handler("shutdown", hot_reload.shutdown)
        templates.env.globals["DEBUG"] = "1"
        templates.env.globals["hot_reload"] = hot_reload
        logger.debug("Debugger (gettrace) detected, UI hot-reload is enabled")
    else:
        logger.debug("In runtime mode ...")


def handle_server_errors(app: FastAPI):
    @app.exception_handler(500)
    async def server_exception_handler(request, exc):
        details = {"title": "Oops! Something went wrong", "message": repr(exc)}
        if isinstance(exc, DialogWindowBase):
            details["title"] = exc.title
            details["message"] = exc.description
            details["status"] = exc.status
        return await http_exception_handler(
            request,
            HTTPException(status_code=500, detail=details),
        )


def _init(
    app: FastAPI = None,
    templates_dir: str = "templates",
    headless=False,
    process_id=None,
    static_dir=None,
) -> FastAPI:
    from supervisely.app.fastapi import available_after_shutdown
    from supervisely.app.content import StateJson, DataJson

    if app is None:
        app = _MainServer().get_server()

    handle_server_errors(app)

    if headless is False:
        if "app_body_padding" not in StateJson():
            StateJson()["app_body_padding"] = "20px"
        Jinja2Templates(directory=[str(Path(__file__).parent.absolute()), templates_dir])
        enable_hot_reload_on_debug(app)

    StateJson()["slyAppShowDialog"] = False
    DataJson()["slyAppDialogTitle"] = ""
    DataJson()["slyAppDialogMessage"] = ""

    app.mount("/sly", create(process_id, headless))

    if headless is False:

        @app.middleware("http")
        async def get_state_from_request(request: Request, call_next):
            await StateJson.from_request(request)
            if not ("application/json" not in request.headers.get("Content-Type", "")):
                # {'command': 'inference_batch_ids', 'context': {}, 'state': {'dataset_id': 49711, 'batch_ids': [3120204], 'settings': None}, 'user_api_key': 'XXX', 'api_token': 'XXX', 'instance_type': None, 'server_address': 'https://dev.supervise.ly'}
                content = await request.json()

                request.state.context = content.get("context")
                request.state.state = content.get("state")
                request.state.api_token = content.get(
                    "api_token",
                    request.state.context.get("apiToken")
                    if request.state.context is not None
                    else None,
                )
                logger.debug(f"middleware request api_token {request.state.api_token}")
                # request.state.server_address = content.get(
                #     "server_address", sly_env.server_address(raise_not_found=False)
                # )
                request.state.server_address = sly_env.server_address(raise_not_found=False)
                logger.debug(f"middleware request server_address {request.state.server_address}")

                logger.debug(f"middleware request context {request.state.context}")

                logger.debug(f"middleware request state {request.state.state}")
                if request.state.server_address is not None and request.state.api_token is not None:
                    request.state.api = Api(request.state.server_address, request.state.api_token)
                else:
                    request.state.api = None

            response = await call_next(request)
            return response

        @app.get("/")
        @available_after_shutdown(app)
        def read_index(request: Request):
            return Jinja2Templates().TemplateResponse("index.html", {"request": request})

        @app.on_event("shutdown")
        def shutdown():
            client = TestClient(app)
            resp = run_sync(client.get("/"))
            assert resp.status_code == 200
            logger.info("Application has been shut down successfully")

        if static_dir is not None:
            app.mount("/static", StaticFiles(directory=static_dir), name="static_files")

    return app


class _MainServer(metaclass=Singleton):
    def __init__(self):
        self._server = FastAPI()

    def get_server(self) -> FastAPI:
        return self._server


class Application(metaclass=Singleton):
    def __init__(self, layout: "Widget" = None, templates_dir: str = None, static_dir: str = None):
        self._favicon = os.environ.get("icon", "https://cdn.supervise.ly/favicon.ico")
        JinjaWidgets().context["__favicon__"] = self._favicon
        JinjaWidgets().context["__no_html_mode__"] = True

        headless = False
        if layout is None and templates_dir is None:
            templates_dir: str = "templates"  # for back compatibility
            headless = True
            logger.info(
                "Both arguments 'layout' and 'templates_dir' are not defined. App is headless (i.e. without UI)",
                extra={"templates_dir": templates_dir},
            )
        if layout is not None and templates_dir is not None:
            raise ValueError(
                "Only one of the arguments has to be defined: 'layout' or 'templates_dir'. 'layout' argument is recommended."
            )
        if layout is not None:
            templates_dir = os.path.join(Path(__file__).parent.absolute(), "templates")
            from supervisely.app.widgets import Identity

            main_layout = Identity(layout, widget_id="__app_main_layout__")
            logger.info(
                "Application is running in no-html mode", extra={"templates_dir": templates_dir}
            )
        else:
            JinjaWidgets().auto_widget_id = False
            JinjaWidgets().context["__no_html_mode__"] = False

        if is_production():
            logger.info("Application is running on Supervisely Platform in production mode")
        else:
            logger.info("Application is running on localhost in development mode")
        self._process_id = os.getpid()
        logger.info(f"Application PID is {self._process_id}")
        self._fastapi: FastAPI = _init(
            app=None,
            templates_dir=templates_dir,
            headless=headless,
            process_id=self._process_id,
            static_dir=static_dir,
        )

    def get_server(self):
        return self._fastapi

    async def __call__(self, scope, receive, send) -> None:
        await self._fastapi.__call__(scope, receive, send)

    def shutdown(self):
        shutdown(self._process_id)


def get_name_from_env(default="Supervisely App"):
    return os.environ.get("APP_NAME", default)



# coding: utf-8

import traceback
import requests
import time

CONNECTION_ERROR = 'Temporary connection error, please wait ...'
AGENT_CONNECTION_ERROR = 'Temporary connection error (agent ping), please wait ...'

REQUEST_FAILED = 'Request has failed. This may be due to connection problems or invalid requests. '
SPECIAL_RECONNECT_ERROR = "Agent should call AgentConnected or AgentPing before attempting any other request"
RETRY_STATUS_CODES = {
    408,  # Request Timeout
    429,  # Too Many Requests
    500,  # Internal Server Error
    502,  # Bad Gateway
    503,  # Service Unavailable
    504,  # Gateway Timeout
    509,  # Bandwidth Limit Exceeded (Apache)
    598,  # Network read timeout error
    599   # Network connect timeout error
}


def process_requests_exception(external_logger, exc, api_method_name, url,
                               verbose=True, swallow_exc=False, sleep_sec=None, response=None, retry_info=None):
    is_connection_error = isinstance(exc, (requests.exceptions.ConnectionError,
                                           requests.exceptions.Timeout,
                                           requests.exceptions.TooManyRedirects,
                                           requests.exceptions.ChunkedEncodingError))

    is_server_retryable_error = isinstance(exc, requests.exceptions.HTTPError) and \
                                hasattr(exc, 'response') and \
                                (exc.response.status_code in RETRY_STATUS_CODES)

    is_need_ping_error = False
    if isinstance(exc, requests.exceptions.HTTPError) and hasattr(exc, 'response') and (exc.response.status_code == 400):
        try:
            server_explanation = exc.response.json()
            is_need_ping_error = (server_explanation.get('error', None) == SPECIAL_RECONNECT_ERROR)
        except (AttributeError, ValueError):
            pass

    if is_connection_error or is_server_retryable_error:
        process_retryable_request(external_logger, exc, api_method_name, url,
                                  CONNECTION_ERROR,
                                  verbose=verbose, swallow_exc=swallow_exc, sleep_sec=sleep_sec, retry_info=retry_info)
    elif is_need_ping_error:
        process_retryable_request(external_logger, exc, api_method_name, url,
                                  AGENT_CONNECTION_ERROR,
                                  verbose=verbose, swallow_exc=swallow_exc, sleep_sec=sleep_sec, retry_info=retry_info)
    elif isinstance(exc, requests.exceptions.HTTPError):
        process_invalid_request(external_logger, exc, response, verbose)
    else:
        process_unhandled_request(external_logger, exc)


def process_retryable_request(external_logger, exc, api_method_name, url, user_message,
                              verbose=True, swallow_exc=False, sleep_sec=None, retry_info=None):
    if retry_info is not None:
        retry_idx = retry_info["retry_idx"]
        retry_limit = retry_info["retry_limit"]
        user_message = "{}:  Retrying ({}/{}).".format(user_message, retry_idx, retry_limit)
    if verbose:
        external_logger.warn(user_message, extra={'method': api_method_name, 'url': url, 'details': str(exc)})

    if sleep_sec is not None:
        time.sleep(sleep_sec)

    if not swallow_exc:
        raise exc


def process_invalid_request(external_logger, exc, response, verbose=True):
    if verbose:
        external_logger.warn(REQUEST_FAILED, extra={'reason': response.content.decode('utf-8'),
                                                    'status_code': response.status_code,
                                                    'url': response.url})
    raise exc


def process_unhandled_request(external_logger, exc):
    exc_str = str(exc)
    external_logger.error(traceback.format_exc(), exc_info=True, extra={'exc_str': exc_str})
    raise RuntimeError(REQUEST_FAILED + 'Last failure: {!r}'.format(exc_str))



from __future__ import annotations
from typing import List, Dict, Optional, Any, Iterator

from supervisely.video_annotation.frame_collection import FrameCollection
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.pointcloud_annotation.pointcloud_episode_frame import PointcloudEpisodeFrame
from supervisely.pointcloud_annotation.pointcloud_object_collection import PointcloudObjectCollection

class PointcloudEpisodeFrameCollection(FrameCollection):
    item_type = PointcloudEpisodeFrame

    def get(self, key: str, default: Optional[Any]=None) -> PointcloudEpisodeFrame:
        return super().get(key, default)

    def __iter__(self) -> Iterator[PointcloudEpisodeFrame]:
        return next(self)

    @classmethod
    def from_json(
        cls, 
        data: List[Dict], 
        objects: PointcloudObjectCollection, 
        frames_count: Optional[int] = None,
        key_id_map: Optional[KeyIdMap] = None
    ) -> PointcloudEpisodeFrameCollection:
        return super().from_json(data, objects, frames_count=frames_count, key_id_map=key_id_map)

    @property
    def figures(self) -> List[PointcloudFigure]:
        return super().figures


from typing import List, Optional, Dict
from supervisely.app import StateJson
from supervisely.app.widgets import Widget


class RadioTabs(Widget):
    class Routes:
        VALUE_CHANGED = "value_changed_cb"

    class RadioTabPane:
        def __init__(
            self,
            title: str,
            content: Widget,
            subtitle: Optional[str] = "",
        ):
            self.title = title
            self.subtitle = subtitle
            self.name = title  # identifier corresponding to the active tab
            self.content = content

    def __init__(
        self,
        titles: List[str],
        contents: List[Widget],
        descriptions: Optional[List[str]] = None,
        widget_id=None,
    ):
        if len(titles) != len(contents):
            raise ValueError(
                "titles length must be equal to contents length in RadioTabs widget."
            )
        if len(titles) > 10:
            raise ValueError("You can specify up to 10 tabs.")
        if descriptions is None:
            descriptions = [""] * len(titles)
        else:
            if len(titles) != len(descriptions):
                raise ValueError(
                    "descriptions length must be equal to titles length in RadioTabs widget."
                )
        if len(set(titles)) != len(titles):
            raise ValueError("All of tab labels should be unique.")
        self._items = []
        for title, widget, description in zip(titles, contents, descriptions):
            self._items.append(
                RadioTabs.RadioTabPane(
                    title=title, content=widget, subtitle=description
                )
            )
        self._value = titles[0]
        self._changes_handled = False
        super().__init__(widget_id=widget_id, file_path=__file__)

    def value_changed(self, func):
        route_path = self.get_route_path(RadioTabs.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _value_changed():
            res = self.get_active_tab()
            func(res)

        return _value_changed

    def get_json_data(self) -> Dict:
        return {}

    def get_json_state(self) -> Dict:
        return {"value": self._value}

    def set_active_tab(self, value: str):
        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def get_active_tab(self) -> str:
        return StateJson()[self.widget_id]["value"]



from typing import List, Dict, Union
from supervisely.app.jinja2 import create_env
from supervisely.app.content import DataJson, StateJson
from supervisely.app.widgets import Widget


class RadioTable(Widget):
    class Routes:
        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        columns: List[str],
        rows: List[List[str]],
        subtitles: Union[Dict[str, str], List] = {},  # col_name -> subtitle
        column_formatters: Dict = {},
        widget_id: str = None,
    ):

        self._columns = columns
        self._rows = rows
        if isinstance(subtitles, dict):
            subtitles = [subtitles[col] for col in columns]
        self._subtitles = subtitles
        self._column_formatters = column_formatters

        self._header = []

        for col, subtitle in zip(columns, subtitles):
            self._header.append({"title": col, "subtitle": subtitle})

        self._frows = []

        super().__init__(widget_id=widget_id, file_path=__file__)

        self.rows = rows
        self._changes_handled = False

    def get_json_data(self):
        return {
            "header": self._header,
            "frows": self._frows,
            "raw_rows_data": self.rows,
        }

    def get_json_state(self):
        return {"selectedRow": 0}

    def format_value(self, column_name: str, value):
        fn = self._column_formatters.get(column_name, self.default_formatter)
        return fn(f"data.{self.widget_id}.raw_rows_data[params.ridx][params.vidx]")

    def default_formatter(self, value):
        if value is None:
            return "-"
        return "<div> {{{{ data.{}.raw_rows_data[params.ridx][params.vidx] }}}} </div>".format(
            self.widget_id
        )

    def _update_frows(self):
        self._frows = []
        for idx, row in enumerate(self._rows):
            if len(row) != len(self.columns):
                raise ValueError(
                    f"Row #{idx} length is {len(row)} != number of columns ({len(self.columns)})"
                )
            frow = []
            for col, val in zip(self.columns, row):
                frow.append(self.format_value(col, val))
            self._frows.append(frow)

    def get_selected_row(self, state=StateJson()):
        widget_actual_state = state[self.widget_id]
        widget_actual_data = DataJson()[self.widget_id]
        if widget_actual_state is not None and widget_actual_data is not None:
            selected_row_index = widget_actual_state["selectedRow"]
            return self.rows[selected_row_index]

    def get_selected_row_index(self, state=StateJson()):
        widget_actual_state = state[self.widget_id]
        widget_actual_data = DataJson()[self.widget_id]
        if widget_actual_state is not None and widget_actual_data is not None:
            return widget_actual_state["selectedRow"]

    def value_changed(self, func):
        route_path = self.get_route_path(RadioTable.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True
        print(self._changes_handled)

        @server.post(route_path)
        def _value_changed():
            res = self.get_selected_row()
            func(res)

        return _value_changed

    @property
    def columns(self) -> List[str]:
        return self._columns

    @property
    def subtitles(self) -> List[str]:
        return self._subtitles

    def set_columns(
        self, columns: List[str], subtitles: Union[Dict[str, str], List[str]] = {}
    ) -> None:
        self._columns = columns
        if isinstance(subtitles, dict):
            subtitles = [subtitles[col] for col in columns]
        self._subtitles = subtitles
        self._header = []
        for col, subtitle in zip(columns, subtitles):
            self._header.append({"title": col, "subtitle": subtitle})
        DataJson()[self.widget_id]["header"] = self._header
        DataJson().send_changes()
        self.rows = [[] * len(columns)]

    def set_data(
        self,
        columns: List[str],
        rows: List[List[str]],
        subtitles: Union[Dict[str, str], List[str]] = {},
    ) -> None:
        self._columns = columns
        if isinstance(subtitles, dict):
            subtitles = [subtitles[col] for col in columns]
        self._subtitles = subtitles
        self._header = []
        for col, subtitle in zip(columns, subtitles):
            self._header.append({"title": col, "subtitle": subtitle})
        DataJson()[self.widget_id]["header"] = self._header
        DataJson().send_changes()
        self.rows = rows

    @property
    def rows(self):
        return self._rows

    @rows.setter
    def rows(self, value):
        self._rows = value
        self._update_frows()
        DataJson()[self.widget_id]["frows"] = self._frows
        DataJson()[self.widget_id]["raw_rows_data"] = self._rows
        DataJson().send_changes()

    def select_row(self, row_index):
        if row_index < 0 or row_index >= len(self._rows) - 1:
            raise ValueError(f'Row with index "{row_index}" does not exist')
        StateJson()[self.widget_id]["selectedRow"] = row_index
        StateJson().send_changes()
